@@ -39,6 +39,13 @@ void CnxnInfo_init()
 
 static PyObject* GetHash(PyObject* p)
 {
+#if PY_MAJOR_VERSION >= 3
+    Object bytes(PyUnicode_EncodeUTF8(PyUnicode_AS_UNICODE(p), PyUnicode_GET_SIZE(p), 0));
+    if (!bytes)
+        return 0;
+    p = bytes.Get();
+#endif
+
     if (hashlib)
     {
         Object hash(PyObject_CallMethod(hashlib, "new", "s", "sha1"));
@@ -65,6 +72,9 @@ static PyObject* GetHash(PyObject* p)
 
 static PyObject* CnxnInfo_New(Connection* cnxn)
 {
+#ifdef _MSC_VER
+#pragma warning(disable : 4365)
+#endif
     CnxnInfo* p = PyObject_NEW(CnxnInfo, &CnxnInfoType);
     if (!p)
         return 0;
@@ -104,21 +114,46 @@ static PyObject* CnxnInfo_New(Connection* cnxn)
         p->supports_describeparam = szYN[0] == 'Y';
     }
 
-    // What is the datetime precision?  This unfortunately requires a cursor (HSTMT).
+    // These defaults are tiny, but are necessary for Access.
+    p->varchar_maxlength = 255;
+    p->wvarchar_maxlength = 255;
+    p->binary_maxlength  = 510;
 
     HSTMT hstmt = 0;
     if (SQL_SUCCEEDED(SQLAllocHandle(SQL_HANDLE_STMT, cnxn->hdbc, &hstmt)))
     {
+        SQLINTEGER columnsize;
         if (SQL_SUCCEEDED(SQLGetTypeInfo(hstmt, SQL_TYPE_TIMESTAMP)) && SQL_SUCCEEDED(SQLFetch(hstmt)))
         {
-            SQLINTEGER columnsize;
             if (SQL_SUCCEEDED(SQLGetData(hstmt, 3, SQL_INTEGER, &columnsize, sizeof(columnsize), 0)))
-            {
-                p->datetime_precision = columnsize;
-            }
+                p->datetime_precision = (int)columnsize;
+
+            SQLFreeStmt(hstmt, SQL_CLOSE);
         }
 
-        SQLFreeStmt(hstmt, SQL_CLOSE);
+        if (SQL_SUCCEEDED(SQLGetTypeInfo(hstmt, SQL_VARCHAR)) && SQL_SUCCEEDED(SQLFetch(hstmt)))
+        {
+            if (SQL_SUCCEEDED(SQLGetData(hstmt, 3, SQL_INTEGER, &columnsize, sizeof(columnsize), 0)))
+                p->varchar_maxlength = (int)columnsize;
+        
+            SQLFreeStmt(hstmt, SQL_CLOSE);
+        }
+        
+        if (SQL_SUCCEEDED(SQLGetTypeInfo(hstmt, SQL_WVARCHAR)) && SQL_SUCCEEDED(SQLFetch(hstmt)))
+        {
+            if (SQL_SUCCEEDED(SQLGetData(hstmt, 3, SQL_INTEGER, &columnsize, sizeof(columnsize), 0)))
+                p->wvarchar_maxlength = (int)columnsize;
+        
+            SQLFreeStmt(hstmt, SQL_CLOSE);
+        }
+        
+        if (SQL_SUCCEEDED(SQLGetTypeInfo(hstmt, SQL_BINARY)) && SQL_SUCCEEDED(SQLFetch(hstmt)))
+        {
+            if (SQL_SUCCEEDED(SQLGetData(hstmt, 3, SQL_INTEGER, &columnsize, sizeof(columnsize), 0)))
+                p->binary_maxlength = (int)columnsize;
+        
+            SQLFreeStmt(hstmt, SQL_CLOSE);
+        }
     }
 
     Py_END_ALLOW_THREADS
@@ -157,8 +192,7 @@ PyObject* GetConnectionInfo(PyObject* pConnectionString, Connection* cnxn)
 
 PyTypeObject CnxnInfoType =
 {
-    PyObject_HEAD_INIT(0)
-    0,                                                      // ob_size
+    PyVarObject_HEAD_INIT(0, 0)
     "pyodbc.CnxnInfo",                                      // tp_name
     sizeof(CnxnInfo),                                       // tp_basicsize
     0,                                                      // tp_itemsize
