@@ -16,14 +16,14 @@
 #include "errors.h"
 #include "wrapper.h"
 #include "cnxninfo.h"
+#include "sqlwchar.h"
 
 static char connection_doc[] =
     "Connection objects manage connections to the database.\n"
     "\n"
     "Each manages a single ODBC HDBC.";
 
-static Connection*
-Connection_Validate(PyObject* self)
+static Connection* Connection_Validate(PyObject* self)
 {
     Connection* cnxn;
     
@@ -44,7 +44,7 @@ Connection_Validate(PyObject* self)
     return cnxn;
 }
 
-static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi)
+static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeout)
 {
     // This should have been checked by the global connect function.
     I(PyString_Check(pConnectString) || PyUnicode_Check(pConnectString));
@@ -69,45 +69,44 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi)
 
     SQLRETURN ret;
 
+    if (timeout > 0)
+    {
+        Py_BEGIN_ALLOW_THREADS
+        ret = SQLSetConnectAttr(hdbc, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)timeout, SQL_IS_UINTEGER);
+        Py_END_ALLOW_THREADS
+        if (!SQL_SUCCEEDED(ret))
+            RaiseErrorFromHandle("SQLSetConnectAttr(SQL_ATTR_LOGIN_TIMEOUT)", hdbc, SQL_NULL_HANDLE);
+    }
+
     if (!fAnsi)
     {
-        SQLWCHAR szConnectW[cchMax];
-        if (PyUnicode_Check(pConnectString))
-        {
-            Py_UNICODE* p = PyUnicode_AS_UNICODE(pConnectString);
-            for (int i = 0, c = PyUnicode_GET_SIZE(pConnectString); i <= c; i++)
-                szConnectW[i] = (wchar_t)p[i];
-        }
-        else
-        {
-            const char* p = PyString_AS_STRING(pConnectString);
-            for (int i = 0, c = PyString_GET_SIZE(pConnectString); i <= c; i++)
-                szConnectW[i] = (wchar_t)p[i];
-        }
-
+        SQLWChar connectString(pConnectString);
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLDriverConnectW(hdbc, 0, szConnectW, SQL_NTS, 0, 0, 0, SQL_DRIVER_NOPROMPT);
+        ret = SQLDriverConnectW(hdbc, 0, connectString, (SQLSMALLINT)connectString.size(), 0, 0, 0, SQL_DRIVER_NOPROMPT);
         Py_END_ALLOW_THREADS
         if (SQL_SUCCEEDED(ret))
             return true;
 
         // The Unicode function failed.  If the error is that the driver doesn't have a Unicode version (IM001), continue
         // to the ANSI version.
-
-        PyObject* error = GetErrorFromHandle("SQLDriverConnectW", hdbc, SQL_NULL_HANDLE);
-        if (!HasSqlState(error, "IM001"))
-        {
-            RaiseErrorFromException(error);
-            return false;
-        }
-        Py_XDECREF(error);
+        //
+        // I've commented this out since a number of common drivers are returning different errors.  The MySQL 5
+        // driver, for example, returns IM002 "Data source name not found...".
+        //
+        // PyObject* error = GetErrorFromHandle("SQLDriverConnectW", hdbc, SQL_NULL_HANDLE);
+        // if (!HasSqlState(error, "IM001"))
+        // {
+        //     RaiseErrorFromException(error);
+        //     return false;
+        // }
+        // Py_XDECREF(error);
     }
         
     SQLCHAR szConnect[cchMax];
     if (PyUnicode_Check(pConnectString))
     {
         Py_UNICODE* p = PyUnicode_AS_UNICODE(pConnectString);
-        for (int i = 0, c = PyUnicode_GET_SIZE(pConnectString); i <= c; i++)
+        for (Py_ssize_t i = 0, c = PyUnicode_GET_SIZE(pConnectString); i <= c; i++)
         {
             if (p[i] > 0xFF)
             {
@@ -115,13 +114,18 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi)
                                 "not have a Unicode connect function");
                 return false;
             }
-            szConnect[i] = (char)p[i];
+            szConnect[i] = (SQLCHAR)p[i];
         }
     }
     else
     {
+#if PY_MAJOR_VERSION < 3
         const char* p = PyString_AS_STRING(pConnectString);
-        memcpy(szConnect, p, PyString_GET_SIZE(pConnectString) + 1);
+        memcpy(szConnect, p, (size_t)(PyString_GET_SIZE(pConnectString) + 1));
+#else
+        PyErr_SetString(PyExc_TypeError, "Connection strings must be Unicode");
+        return false;
+#endif
     }
 
     Py_BEGIN_ALLOW_THREADS
@@ -136,7 +140,7 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi)
 }
 
 
-PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, bool fUnicodeResults)
+PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, bool fUnicodeResults, long timeout, bool fReadOnly)
 {
     // pConnectString
     //   A string or unicode object.  (This must be checked by the caller.)
@@ -159,7 +163,7 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLAllocHandle", SQL_NULL_HANDLE, SQL_NULL_HANDLE);
 
-    if (!Connect(pConnectString, hdbc, fAnsi))
+    if (!Connect(pConnectString, hdbc, fAnsi, timeout))
     {
         // Connect has already set an exception.
         Py_BEGIN_ALLOW_THREADS
@@ -174,7 +178,13 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
 
     // Set all variables to something valid, so we don't crash in dealloc if this function fails.
 
+#ifdef _MSC_VER
+#pragma warning(disable : 4365)
+#endif
     Connection* cnxn = PyObject_NEW(Connection, &ConnectionType);
+#ifdef _MSC_VER
+#pragma warning(default : 4365)
+#endif
 
     if (cnxn == 0)
     {
@@ -189,6 +199,9 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     cnxn->searchescape    = 0;
     cnxn->timeout         = 0;
     cnxn->unicode_results = fUnicodeResults;
+    cnxn->conv_count      = 0;
+    cnxn->conv_types      = 0;
+    cnxn->conv_funcs      = 0;
 
     //
     // Initialize autocommit mode.
@@ -213,9 +226,22 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
         }
     }
     
-#ifdef TRACE_ALL
-    printf("cnxn.new cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
-#endif
+    if (fReadOnly)
+    {
+        SQLRETURN ret;
+        Py_BEGIN_ALLOW_THREADS
+        ret = SQLSetConnectAttr(cnxn->hdbc, SQL_ATTR_ACCESS_MODE, (SQLPOINTER)SQL_MODE_READ_ONLY, 0);
+        Py_END_ALLOW_THREADS
+
+        if (!SQL_SUCCEEDED(ret))
+        {
+            RaiseErrorFromHandle("SQLSetConnnectAttr(SQL_ATTR_ACCESS_MODE)", cnxn->hdbc, SQL_NULL_HANDLE);
+            Py_DECREF(cnxn);
+            return 0;
+        }
+    }
+
+    TRACE("cnxn.new cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
 
     //
     // Gather connection-level information we'll need later.
@@ -234,24 +260,55 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     cnxn->odbc_minor             = p->odbc_minor;
     cnxn->supports_describeparam = p->supports_describeparam;
     cnxn->datetime_precision     = p->datetime_precision;
+    cnxn->varchar_maxlength      = p->varchar_maxlength;
+    cnxn->wvarchar_maxlength     = p->wvarchar_maxlength;
+    cnxn->binary_maxlength       = p->binary_maxlength;
 
     return reinterpret_cast<PyObject*>(cnxn);
 }
 
+static void _clear_conv(Connection* cnxn)
+{
+    if (cnxn->conv_count != 0)
+    {
+        pyodbc_free(cnxn->conv_types);
+        cnxn->conv_types = 0;
+        
+        for (int i = 0; i < cnxn->conv_count; i++)
+            Py_XDECREF(cnxn->conv_funcs[i]);
+        pyodbc_free(cnxn->conv_funcs);
+        cnxn->conv_funcs = 0;
 
-static int
-Connection_clear(Connection* cnxn)
+        cnxn->conv_count = 0;
+    }
+}
+
+static char conv_clear_doc[] =
+    "clear_output_converters() --> None\n\n"
+    "Remove all output converter functions.";
+
+static PyObject* Connection_conv_clear(PyObject* self, PyObject* args)
+{
+    UNUSED(args);
+    
+    Connection* cnxn = (Connection*)self;
+    _clear_conv(cnxn);
+    Py_RETURN_NONE;
+}
+
+static int Connection_clear(PyObject* self)
 {
     // Internal method for closing the connection.  (Not called close so it isn't confused with the external close
     // method.)
 
+    Connection* cnxn = (Connection*)self;
+
     if (cnxn->hdbc != SQL_NULL_HANDLE)
     {
         // REVIEW: Release threads? (But make sure you zero out hdbc *first*!
 
-#ifdef TRACE_ALL
-        printf("cnxn.clear cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
-#endif
+        TRACE("cnxn.clear cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
+
         Py_BEGIN_ALLOW_THREADS
         if (cnxn->nAutoCommit == SQL_AUTOCOMMIT_OFF)
             SQLEndTran(SQL_HANDLE_DBC, cnxn->hdbc, SQL_ROLLBACK);
@@ -266,14 +323,14 @@ Connection_clear(Connection* cnxn)
     Py_XDECREF(cnxn->searchescape);
     cnxn->searchescape = 0;
     
+    _clear_conv(cnxn);
+
     return 0;
 }
 
-static void
-Connection_dealloc(PyObject* self)
+static void Connection_dealloc(PyObject* self)
 {
-    Connection* cnxn = (Connection*)self;
-    Connection_clear(cnxn);
+    Connection_clear(self);
     PyObject_Del(self);
 }
 
@@ -287,8 +344,7 @@ static char close_doc[] =
     "Note that closing a connection without committing the changes first will cause\n"
     "an implicit rollback to be performed.";
     
-static PyObject*
-Connection_close(PyObject* self, PyObject* args)
+static PyObject* Connection_close(PyObject* self, PyObject* args)
 {
     UNUSED(args);
     
@@ -296,13 +352,12 @@ Connection_close(PyObject* self, PyObject* args)
     if (!cnxn)
         return 0;
 
-    Connection_clear(cnxn);
+    Connection_clear(self);
 
     Py_RETURN_NONE;
 }
 
-static PyObject*
-Connection_cursor(PyObject* self, PyObject* args)
+static PyObject* Connection_cursor(PyObject* self, PyObject* args)
 {
     UNUSED(args);
     
@@ -313,8 +368,7 @@ Connection_cursor(PyObject* self, PyObject* args)
     return (PyObject*)Cursor_New(cnxn);
 }
 
-static PyObject*
-Connection_execute(PyObject* self, PyObject* args)
+static PyObject* Connection_execute(PyObject* self, PyObject* args)
 {
     PyObject* result = 0;
 
@@ -494,15 +548,14 @@ static const GetInfoType aInfoTypes[] = {
     { SQL_XOPEN_CLI_YEAR, GI_STRING },
 };
 
-static PyObject*
-Connection_getinfo(PyObject* self, PyObject* args)
+static PyObject* Connection_getinfo(PyObject* self, PyObject* args)
 {
     Connection* cnxn = Connection_Validate(self);
     if (!cnxn)
         return 0;
 
-    SQLUSMALLINT infotype;
-    if (!PyArg_ParseTuple(args, "l", &infotype))
+    unsigned long infotype;
+    if (!PyArg_ParseTuple(args, "k", &infotype))
         return 0;
 
     unsigned int i = 0;
@@ -520,7 +573,7 @@ Connection_getinfo(PyObject* self, PyObject* args)
 
     SQLRETURN ret;
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetInfo(cnxn->hdbc, infotype, szBuffer, sizeof(szBuffer), &cch);
+    ret = SQLGetInfo(cnxn->hdbc, (SQLUSMALLINT)infotype, szBuffer, sizeof(szBuffer), &cch);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
     {
@@ -544,10 +597,14 @@ Connection_getinfo(PyObject* self, PyObject* args)
     case GI_UINTEGER:
     {
         SQLUINTEGER n = *(SQLUINTEGER*)szBuffer; // Does this work on PPC or do we need a union?
+#if PY_MAJOR_VERSION >= 3
+        result = PyLong_FromLong((long)n);
+#else
         if (n <= (SQLUINTEGER)PyInt_GetMax())
             result = PyInt_FromLong((long)n);
         else
             result = PyLong_FromUnsignedLong(n);
+#endif
         break;
     }
     
@@ -559,9 +616,28 @@ Connection_getinfo(PyObject* self, PyObject* args)
     return result;
 }
 
+PyObject* Connection_endtrans(Connection* cnxn, SQLSMALLINT type)
+{
+    // If called from Cursor.commit, it is possible that `cnxn` is deleted by another thread when we release them
+    // below.  (The cursor has had its reference incremented by the method it is calling, but nothing has incremented
+    // the connections count.  We could, but we really only need the HDBC.)
+    HDBC hdbc = cnxn->hdbc;
 
-static PyObject*
-Connection_endtrans(PyObject* self, PyObject* args, SQLSMALLINT type)
+    SQLRETURN ret;
+    Py_BEGIN_ALLOW_THREADS
+    ret = SQLEndTran(SQL_HANDLE_DBC, hdbc, type);
+    Py_END_ALLOW_THREADS
+
+    if (!SQL_SUCCEEDED(ret))
+    {
+        RaiseErrorFromHandle("SQLEndTran", hdbc, SQL_NULL_HANDLE);
+        return 0;
+    }
+
+    Py_RETURN_NONE;
+}
+
+static PyObject* Connection_commit(PyObject* self, PyObject* args)
 {
     UNUSED(args);
     
@@ -569,47 +645,35 @@ Connection_endtrans(PyObject* self, PyObject* args, SQLSMALLINT type)
     if (!cnxn)
         return 0;
     
-#ifdef TRACE_ALL
-    printf("%s: cnxn=%p hdbc=%d\n", (type == SQL_COMMIT) ? "commit" : "rollback", cnxn, cnxn->hdbc);
-#endif
+    TRACE("commit: cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
 
-    SQLRETURN ret;
-    Py_BEGIN_ALLOW_THREADS
-    ret = SQLEndTran(SQL_HANDLE_DBC, cnxn->hdbc, type);
-    Py_END_ALLOW_THREADS
-    if (!SQL_SUCCEEDED(ret))
-    {
-        RaiseErrorFromHandle("SQLEndTran", cnxn->hdbc, SQL_NULL_HANDLE);
+    return Connection_endtrans(cnxn, SQL_COMMIT);
+}
+
+static PyObject* Connection_rollback(PyObject* self, PyObject* args)
+{
+    UNUSED(args);
+    
+    Connection* cnxn = Connection_Validate(self);
+    if (!cnxn)
         return 0;
-    }
+    
+    TRACE("rollback: cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
 
-    Py_RETURN_NONE;
-}
-
-static PyObject*
-Connection_commit(PyObject* self, PyObject* args)
-{
-    return Connection_endtrans(self, args, SQL_COMMIT);
-}
-
-static PyObject*
-Connection_rollback(PyObject* self, PyObject* args)
-{
-    return Connection_endtrans(self, args, SQL_ROLLBACK);
+    return Connection_endtrans(cnxn, SQL_ROLLBACK);
 }
 
 static char cursor_doc[] = 
-    "Return a new Cursor Object using the connection.";
+    "Return a new Cursor object using the connection.";
     
 static char execute_doc[] =
-    "execute(sql, [params]) --> None | Cursor | count\n" \
-    "\n" \
-    "Creates a new Cursor object, calls its execute method, and returns its return\n" \
-    "value.  See Cursor.execute for a description of the parameter formats and\n" \
-    "return values.\n" \
-    "\n" \
-    "This is a convenience method that is not part of the DB API.  Since a new\n" \
-    "Cursor is allocated by each call, this should not be used if more than one SQL\n" \
+    "execute(sql, [params]) --> Cursor\n"
+    "\n"
+    "Create a new Cursor object, call its execute method, and return it.  See\n"
+    "Cursor.execute for more details.\n"
+    "\n"
+    "This is a convenience method that is not part of the DB API.  Since a new\n"
+    "Cursor is allocated by each call, this should not be used if more than one SQL\n"
     "statement needs to be executed.";
 
 static char commit_doc[] =
@@ -624,8 +688,7 @@ static char getinfo_doc[] =
     "Calls SQLGetInfo, passing `type`, and returns the result formatted as a Python object.";
 
 
-PyObject*
-Connection_getautocommit(PyObject* self, void* closure)
+PyObject* Connection_getautocommit(PyObject* self, void* closure)
 {
     UNUSED(closure);
 
@@ -638,8 +701,7 @@ Connection_getautocommit(PyObject* self, void* closure)
     return result;
 }
 
-static int
-Connection_setautocommit(PyObject* self, PyObject* value, void* closure)
+static int Connection_setautocommit(PyObject* self, PyObject* value, void* closure)
 {
     UNUSED(closure);
 
@@ -653,7 +715,7 @@ Connection_setautocommit(PyObject* self, PyObject* value, void* closure)
         return -1;
     }
     
-    int nAutoCommit = PyObject_IsTrue(value) ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF;
+    uintptr_t nAutoCommit = PyObject_IsTrue(value) ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF;
     SQLRETURN ret;
     Py_BEGIN_ALLOW_THREADS
     ret = SQLSetConnectAttr(cnxn->hdbc, SQL_ATTR_AUTOCOMMIT, (SQLPOINTER)nAutoCommit, SQL_IS_UINTEGER);
@@ -670,33 +732,33 @@ Connection_setautocommit(PyObject* self, PyObject* value, void* closure)
 }
 
 
-PyObject*
-Connection_getsearchescape(Connection* self, void* closure)
+static PyObject* Connection_getsearchescape(PyObject* self, void* closure)
 {
     UNUSED(closure);
     
-    if (!self->searchescape)
+    Connection* cnxn = (Connection*)self;
+
+    if (!cnxn->searchescape)
     {
         char sz[8] = { 0 };
         SQLSMALLINT cch = 0;
 
         SQLRETURN ret;
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLGetInfo(self->hdbc, SQL_SEARCH_PATTERN_ESCAPE, &sz, _countof(sz), &cch);
+        ret = SQLGetInfo(cnxn->hdbc, SQL_SEARCH_PATTERN_ESCAPE, &sz, _countof(sz), &cch);
         Py_END_ALLOW_THREADS
         if (!SQL_SUCCEEDED(ret))
-            return RaiseErrorFromHandle("SQLGetInfo", self->hdbc, SQL_NULL_HANDLE);
+            return RaiseErrorFromHandle("SQLGetInfo", cnxn->hdbc, SQL_NULL_HANDLE);
 
-        self->searchescape = PyString_FromStringAndSize(sz, (Py_ssize_t)cch);
+        cnxn->searchescape = PyString_FromStringAndSize(sz, (Py_ssize_t)cch);
     }
 
-    Py_INCREF(self->searchescape);
-    return self->searchescape;
+    Py_INCREF(cnxn->searchescape);
+    return cnxn->searchescape;
 }
 
 
-static PyObject*
-Connection_gettimeout(PyObject* self, void* closure)
+static PyObject* Connection_gettimeout(PyObject* self, void* closure)
 {
     UNUSED(closure);
 
@@ -707,8 +769,7 @@ Connection_gettimeout(PyObject* self, void* closure)
     return PyInt_FromLong(cnxn->timeout);
 }
 
-static int
-Connection_settimeout(PyObject* self, PyObject* value, void* closure)
+static int Connection_settimeout(PyObject* self, PyObject* value, void* closure)
 {
     UNUSED(closure);
 
@@ -721,7 +782,7 @@ Connection_settimeout(PyObject* self, PyObject* value, void* closure)
         PyErr_SetString(PyExc_TypeError, "Cannot delete the timeout attribute.");
         return -1;
     }
-    int timeout = PyInt_AsLong(value);
+    intptr_t timeout = PyInt_AsLong(value);
     if (timeout == -1 && PyErr_Occurred())
         return -1;
     if (timeout < 0)
@@ -745,14 +806,130 @@ Connection_settimeout(PyObject* self, PyObject* value, void* closure)
     return 0;
 }
 
+static bool _add_converter(PyObject* self, SQLSMALLINT sqltype, PyObject* func)
+{
+    Connection* cnxn = (Connection*)self;
+
+    if (cnxn->conv_count)
+    {
+        // If the sqltype is already registered, replace the old conversion function with the new.
+        for (int i = 0; i < cnxn->conv_count; i++)
+        {
+            if (cnxn->conv_types[i] == sqltype)
+            {
+                Py_XDECREF(cnxn->conv_funcs[i]);
+                cnxn->conv_funcs[i] = func;
+                Py_INCREF(func);
+                return true;
+            }
+        }
+    }
+    
+    int          oldcount = cnxn->conv_count;
+    SQLSMALLINT* oldtypes = cnxn->conv_types;
+    PyObject**   oldfuncs = cnxn->conv_funcs;
+
+    int          newcount = oldcount + 1;
+    SQLSMALLINT* newtypes = (SQLSMALLINT*)pyodbc_malloc(sizeof(SQLSMALLINT) * newcount);
+    PyObject**   newfuncs = (PyObject**)pyodbc_malloc(sizeof(PyObject*) * newcount);
+
+    if (newtypes == 0 || newfuncs == 0)
+    {
+        if (newtypes)
+            pyodbc_free(newtypes);
+        if (newfuncs)
+            pyodbc_free(newfuncs);
+        PyErr_NoMemory();
+        return false;
+    }
+
+    newtypes[0] = sqltype;
+    newfuncs[0] = func;
+    Py_INCREF(func);
+
+    cnxn->conv_count = newcount;
+    cnxn->conv_types = newtypes;
+    cnxn->conv_funcs = newfuncs;
+
+    if (oldcount != 0)
+    {
+        // copy old items
+        memcpy(&newtypes[1], oldtypes, sizeof(int) * oldcount);
+        memcpy(&newfuncs[1], oldfuncs, sizeof(PyObject*) * oldcount);
+
+        pyodbc_free(oldtypes);
+        pyodbc_free(oldfuncs);
+    }
+
+    return true;
+}
+
+static char conv_add_doc[] =
+    "add_output_converter(sqltype, func) --> None\n"
+    "\n"
+    "Register an output converter function that will be called whenever a value with\n"
+    "the given SQL type is read from the database.\n"
+    "\n"
+    "sqltype\n"
+    "  The integer SQL type value to convert, which can be one of the defined\n"
+    "  standard constants (e.g. pyodbc.SQL_VARCHAR) or a database-specific value\n"
+    "  (e.g. -151 for the SQL Server 2008 geometry data type).\n"
+    "\n"
+    "func\n"
+    "  The converter function which will be called with a single parameter, the\n"
+    "  value, and should return the converted value.  If the value is NULL, the\n"
+    "  parameter will be None.  Otherwise it will be a Python string.";
+
+
+static PyObject* Connection_conv_add(PyObject* self, PyObject* args)
+{
+    int sqltype;
+    PyObject* func;
+    if (!PyArg_ParseTuple(args, "iO", &sqltype, &func))
+        return 0;
+
+    if (!_add_converter(self, (SQLSMALLINT)sqltype, func))
+        return 0;
+
+    Py_RETURN_NONE;
+}
+
+static char enter_doc[] = "__enter__() -> self.";
+static PyObject* Connection_enter(PyObject* self, PyObject* args)
+{
+    UNUSED(args);
+    Py_INCREF(self);
+    return self;
+}
+
+static char exit_doc[] = "__exit__(*excinfo) -> None.  Closes the connection.";
+static PyObject* Connection_exit(PyObject* self, PyObject* args)
+{
+    Connection* cnxn = (Connection*)self;
+
+    // If an error has occurred, `args` will be a tuple of 3 values.  Otherwise it will be a tuple of 3 `None`s.
+    I(PyTuple_Check(args));
+
+    if (cnxn->nAutoCommit == SQL_AUTOCOMMIT_OFF && PyTuple_GetItem(args, 0) == Py_None)
+        SQLEndTran(SQL_HANDLE_DBC, cnxn->hdbc, SQL_COMMIT);
+
+    Py_RETURN_NONE;
+}
+
+
 static struct PyMethodDef Connection_methods[] =
 {
-    { "cursor",        (PyCFunction)Connection_cursor,   METH_NOARGS,  cursor_doc    },
-    { "close",         (PyCFunction)Connection_close,    METH_NOARGS,  close_doc     },
-    { "execute",       (PyCFunction)Connection_execute,  METH_VARARGS, execute_doc   },
-    { "commit",        (PyCFunction)Connection_commit,   METH_NOARGS,  commit_doc    },
-    { "rollback",      (PyCFunction)Connection_rollback, METH_NOARGS,  rollback_doc  },
-    { "getinfo",       (PyCFunction)Connection_getinfo,  METH_VARARGS, getinfo_doc   },
+    { "cursor",                  Connection_cursor,          METH_NOARGS,  cursor_doc     },
+    { "close",                   Connection_close,           METH_NOARGS,  close_doc      },
+    { "execute",                 Connection_execute,         METH_VARARGS, execute_doc    },
+    { "commit",                  Connection_commit,          METH_NOARGS,  commit_doc     },
+    { "rollback",                Connection_rollback,        METH_NOARGS,  rollback_doc   },
+    { "getinfo",                 Connection_getinfo,         METH_VARARGS, getinfo_doc    },
+    { "add_output_converter",    Connection_conv_add,        METH_VARARGS, conv_add_doc   },
+    { "clear_output_converters", Connection_conv_clear,      METH_NOARGS,  conv_clear_doc },
+    { "__enter__",               Connection_enter,           METH_NOARGS,  enter_doc      },
+    { "__exit__",                Connection_exit,            METH_VARARGS, exit_doc       },
+    
     { 0, 0, 0, 0 }
 };
 
@@ -769,50 +946,49 @@ static PyGetSetDef Connection_getseters[] = {
 
 PyTypeObject ConnectionType =
 {
-    PyObject_HEAD_INIT(0)
-    0,                                                      // ob_size
-    "pyodbc.Connection",                                    // tp_name
-    sizeof(Connection),                                     // tp_basicsize
-    0,                                                      // tp_itemsize
-    (destructor)Connection_dealloc,                         // destructor tp_dealloc
-    0,                                                      // tp_print
-    0,                                                      // tp_getattr
-    0,                                                      // tp_setattr
-    0,                                                      // tp_compare
-    0,                                                      // tp_repr
-    0,                                                      // tp_as_number
-    0,                                                      // tp_as_sequence
-    0,                                                      // tp_as_mapping
-    0,                                                      // tp_hash
-    0,                                                      // tp_call
-    0,                                                      // tp_str
-    0,                                                      // tp_getattro
-    0,                                                      // tp_setattro
-    0,                                                      // tp_as_buffer
-    Py_TPFLAGS_DEFAULT,                                     // tp_flags
-    connection_doc,                                         // tp_doc
-    0,                                                      // tp_traverse
-    0,                                                      // tp_clear
-    0,                                                      // tp_richcompare
-    0,                                                      // tp_weaklistoffset
-    0,                                                      // tp_iter
-    0,                                                      // tp_iternext
-    Connection_methods,                                     // tp_methods
-    0,                                                      // tp_members
-    Connection_getseters,                                   // tp_getset
-    0,                                                      // tp_base
-    0,                                                      // tp_dict
-    0,                                                      // tp_descr_get
-    0,                                                      // tp_descr_set
-    0,                                                      // tp_dictoffset
-    0,                                                      // tp_init
-    0,                                                      // tp_alloc
-    0,                                                      // tp_new
-    0,                                                      // tp_free
-    0,                                                      // tp_is_gc
-    0,                                                      // tp_bases
-    0,                                                      // tp_mro
-    0,                                                      // tp_cache
-    0,                                                      // tp_subclasses
-    0,                                                      // tp_weaklist
+    PyVarObject_HEAD_INIT(0, 0)
+    "pyodbc.Connection",        // tp_name
+    sizeof(Connection),         // tp_basicsize
+    0,                          // tp_itemsize
+    Connection_dealloc,         // destructor tp_dealloc
+    0,                          // tp_print
+    0,                          // tp_getattr
+    0,                          // tp_setattr
+    0,                          // tp_compare
+    0,                          // tp_repr
+    0,                          // tp_as_number
+    0,                          // tp_as_sequence
+    0,                          // tp_as_mapping
+    0,                          // tp_hash
+    0,                          // tp_call
+    0,                          // tp_str
+    0,                          // tp_getattro
+    0,                          // tp_setattro
+    0,                          // tp_as_buffer
+    Py_TPFLAGS_DEFAULT,         // tp_flags
+    connection_doc,             // tp_doc
+    0,                          // tp_traverse
+    0,                          // tp_clear
+    0,                          // tp_richcompare
+    0,                          // tp_weaklistoffset
+    0,                          // tp_iter
+    0,                          // tp_iternext
+    Connection_methods,         // tp_methods
+    0,                          // tp_members
+    Connection_getseters,       // tp_getset
+    0,                          // tp_base
+    0,                          // tp_dict
+    0,                          // tp_descr_get
+    0,                          // tp_descr_set
+    0,                          // tp_dictoffset
+    0,                          // tp_init
+    0,                          // tp_alloc
+    0,                          // tp_new
+    0,                          // tp_free
+    0,                          // tp_is_gc
+    0,                          // tp_bases
+    0,                          // tp_mro
+    0,                          // tp_cache
+    0,                          // tp_subclasses
+    0,                          // tp_weaklist
 };
