@@ -24,7 +24,7 @@ struct Connection
 	HDBC hdbc;
 
     // Will be SQL_AUTOCOMMIT_ON or SQL_AUTOCOMMIT_OFF.
-    int nAutoCommit;
+    uintptr_t nAutoCommit;
 
     // The ODBC version the driver supports, from SQLGetInfo(DRIVER_ODBC_VER).  This is set after connecting.
     char odbc_major;
@@ -43,17 +43,39 @@ struct Connection
     // If true, then the strings in the rows are returned as unicode objects.
     bool unicode_results;
 
-     // The connection timeout in seconds.
-     int timeout;
+    // The connection timeout in seconds.
+    intptr_t timeout;
+
+    // These are copied from cnxn info for performance and convenience.
+
+    int varchar_maxlength;
+    int wvarchar_maxlength;
+    int binary_maxlength;
+
+    // Output conversions.  Maps from SQL type in conv_types to the converter function in conv_funcs.
+    //
+    // If conv_count is zero, conv_types and conv_funcs will also be zero.
+    //
+    // pyodbc uses this manual mapping for speed and portability.  The STL collection classes use the new operator and
+    // throw exceptions when out of memory.  pyodbc does not use any exceptions.
+
+    int conv_count;             // how many items are in conv_types and conv_funcs.
+    SQLSMALLINT* conv_types;            // array of SQL_TYPEs to convert
+    PyObject** conv_funcs;      // array of Python functions
 };
 
 #define Connection_Check(op) PyObject_TypeCheck(op, &ConnectionType)
-#define Connection_CheckExact(op) ((op)->ob_type == &ConnectionType)
+#define Connection_CheckExact(op) (Py_TYPE(op) == &ConnectionType)
 
 /*
  * Used by the module's connect function to create new connection objects.  If unable to connect to the database, an
  * exception is set and zero is returned.
  */
-PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, bool fUnicodeResults);
+PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, bool fUnicodeResults, long timeout, bool fReadOnly);
+
+/*
+ * Used by the Cursor to implement commit and rollback.
+ */
+PyObject* Connection_endtrans(Connection* cnxn, SQLSMALLINT type);
 
 #endif
