@@ -52,8 +52,12 @@ bool HasSqlState(HSTMT hstmt, const char* szSqlState);
 inline PyObject* RaiseErrorFromException(PyObject* pError)
 {
     // PyExceptionInstance_Class doesn't exist in 2.4
-	PyObject* cls = (PyObject*)((PyInstance_Check(pError) ? (PyObject*)((PyInstanceObject*)pError)->in_class : (PyObject*)(pError->ob_type)));
+#if PY_MAJOR_VERSION >= 3
+    PyErr_SetObject((PyObject*)Py_TYPE(pError), pError);
+#else
+	PyObject* cls = (PyObject*)((PyInstance_Check(pError) ? (PyObject*)((PyInstanceObject*)pError)->in_class : (PyObject*)(Py_TYPE(pError))));
     PyErr_SetObject(cls, pError);
+#endif
     return 0;
 }
 
