@@ -8,10 +8,13 @@
 #include "connection.h"
 #include "errors.h"
 #include "dbspecific.h"
+#include "sqlwchar.h"
+#include "wrapper.h"
+#include <datetime.h>
 
 void GetData_init()
 {
-    PyDateTime_IMPORT;    
+    PyDateTime_IMPORT;
 }
 
 class DataBuffer
@@ -24,30 +27,32 @@ class DataBuffer
     //
     //   1) Binary, which is a simple array of 8-bit bytes.
     //   2) ANSI text, which is an array of chars with a NULL terminator.
-    //   3) Unicode text, which is an array of wchar_ts with a NULL terminator.
+    //   3) Unicode text, which is an array of SQLWCHARs with a NULL terminator.
     //
-    // When dealing with Unicode, there are two widths we have to be aware of: (1) wchar_t and (2) Py_UNICODE.  If
+    // When dealing with Unicode, there are two widths we have to be aware of: (1) SQLWCHAR and (2) Py_UNICODE.  If
     // these are the same we can use a PyUnicode object so we don't have to allocate our own buffer and then the
     // Unicode object.  If they are not the same (e.g. OS/X where wchar_t-->4 Py_UNICODE-->2) then we need to maintain
-    // our own buffer and pass it to the PyUnicode object later.
+    // our own buffer and pass it to the PyUnicode object later.  Many Linux distros are now using UCS4, so Py_UNICODE
+    // will be larger than SQLWCHAR.
     //
     // To reduce heap fragmentation, we perform the initial read into an array on the stack since we don't know the
-    // length of the data.  If the data doesn't fit, this class then allocates new memory.
+    // length of the data.  If the data doesn't fit, this class then allocates new memory.  If the first read gives us
+    // the length, then we create a Python object of the right size and read into its memory.
 
 private:
     SQLSMALLINT dataType;
 
     char* buffer;
     Py_ssize_t bufferSize;      // How big is the buffer.
-    int bytesUsed;          // How many elements have been read into the buffer?
+    int bytesUsed;              // How many elements have been read into the buffer?
 
-    PyObject* bufferOwner;      // If possible, we bind into a PyString or PyUnicode object.
+    PyObject* bufferOwner;      // If possible, we bind into a PyString, PyUnicode, or PyByteArray object.
     int element_size;           // How wide is each character: ASCII/ANSI -> 1, Unicode -> 2 or 4, binary -> 1
 
     bool usingStack;            // Is buffer pointing to the initial stack buffer?
 
 public:
-    int null_size;              // How much room to add for null terminator: binary -> 0, other -> same as a element_size
+    int null_size;              // How much room, in bytes, to add for null terminator: binary -> 0, other -> same as a element_size
 
     DataBuffer(SQLSMALLINT dataType, char* stackBuffer, SQLLEN stackBufferSize)
     {
@@ -56,7 +61,7 @@ public:
 
         this->dataType = dataType;
 
-        element_size = (dataType == SQL_C_WCHAR) ? sizeof(wchar_t) : sizeof(char);
+        element_size = (int)((dataType == SQL_C_WCHAR)  ? sizeof(SQLWCHAR) : sizeof(char));
         null_size    = (dataType == SQL_C_BINARY) ? 0 : element_size;
 
         buffer        = stackBuffer;
@@ -76,7 +81,7 @@ public:
             }
             else
             {
-                free(buffer);
+                pyodbc_free(buffer);
             }
         }
     }
@@ -98,11 +103,14 @@ public:
     void AddUsed(SQLLEN cbRead)
     {
         I(cbRead <= GetRemaining());
-        bytesUsed += cbRead;
+        bytesUsed += (int)cbRead;
     }
 
     bool AllocateMore(SQLLEN cbAdd)
     {
+        // cbAdd
+        //   The number of bytes (cb --> count of bytes) to add.
+
         if (cbAdd == 0)
             return true;
 
@@ -112,15 +120,25 @@ public:
         {
             // This is the first call and `buffer` points to stack memory.  Allocate a new object and copy the stack
             // data into it.
-            
+
             char* stackBuffer = buffer;
 
-            if (dataType == SQL_C_CHAR || dataType == SQL_C_BINARY)
+            if (dataType == SQL_C_CHAR)
             {
-                bufferOwner = PyString_FromStringAndSize(0, newSize);
-                buffer      = bufferOwner ? PyString_AS_STRING(bufferOwner) : 0;
+                bufferOwner = PyBytes_FromStringAndSize(0, newSize);
+                buffer      = bufferOwner ? PyBytes_AS_STRING(bufferOwner) : 0;
             }
-            else if (sizeof(wchar_t) == Py_UNICODE_SIZE)
+            else if (dataType == SQL_C_BINARY)
+            {
+#if PY_VERSION_HEX >= 0x02060000
+                bufferOwner = PyByteArray_FromStringAndSize(0, newSize);
+                buffer      = bufferOwner ? PyByteArray_AS_STRING(bufferOwner) : 0;
+#else
+                bufferOwner = PyBytes_FromStringAndSize(0, newSize);
+                buffer      = bufferOwner ? PyBytes_AS_STRING(bufferOwner) : 0;
+#endif
+            }
+            else if (sizeof(SQLWCHAR) == Py_UNICODE_SIZE)
             {
                 // Allocate directly into a Unicode object.
                 bufferOwner = PyUnicode_FromUnicode(0, newSize / element_size);
@@ -128,35 +146,44 @@ public:
             }
             else
             {
-                // We're Unicode, but wchar_t and Py_UNICODE don't match, so maintain our own wchar_t buffer.
-                buffer = (char*)malloc(newSize);
+                // We're Unicode, but SQLWCHAR and Py_UNICODE don't match, so maintain our own SQLWCHAR buffer.
+                bufferOwner = 0;
+                buffer      = (char*)pyodbc_malloc((size_t)newSize);
             }
 
-            usingStack = false;
-
             if (buffer == 0)
                 return false;
 
-            memcpy(buffer, stackBuffer, bufferSize);
+            usingStack = false;
+
+            memcpy(buffer, stackBuffer, (size_t)bufferSize);
             bufferSize = newSize;
             return true;
         }
 
-        if (PyString_CheckExact(bufferOwner))
-        {
-            if (_PyString_Resize(&bufferOwner, newSize) == -1)
-                return false;
-            buffer = PyString_AS_STRING(bufferOwner);
-        }
-        else if (PyUnicode_CheckExact(bufferOwner))
+        if (bufferOwner && PyUnicode_CheckExact(bufferOwner))
         {
             if (PyUnicode_Resize(&bufferOwner, newSize / element_size) == -1)
                 return false;
             buffer = (char*)PyUnicode_AsUnicode(bufferOwner);
         }
+#if PY_VERSION_HEX >= 0x02060000
+        else if (bufferOwner && PyByteArray_CheckExact(bufferOwner))
+        {
+            if (PyByteArray_Resize(bufferOwner, newSize) == -1)
+                return false;
+            buffer = PyByteArray_AS_STRING(bufferOwner);
+        }
+#endif
+        else if (bufferOwner && PyBytes_CheckExact(bufferOwner))
+        {
+            if (_PyBytes_Resize(&bufferOwner, newSize) == -1)
+                return false;
+            buffer = PyBytes_AS_STRING(bufferOwner);
+        }
         else
         {
-            char* tmp = (char*)realloc(buffer, newSize);
+            char* tmp = (char*)realloc(buffer, (size_t)newSize);
             if (tmp == 0)
                 return false;
             buffer = tmp;
@@ -176,26 +203,25 @@ public:
 
         if (usingStack)
         {
-            if (dataType == SQL_C_CHAR || dataType == SQL_C_BINARY)
-                return PyString_FromStringAndSize(buffer, bytesUsed);
+            if (dataType == SQL_C_CHAR)
+                return PyBytes_FromStringAndSize(buffer, bytesUsed);
 
-            if (sizeof(wchar_t) == Py_UNICODE_SIZE)
+            if (dataType == SQL_C_BINARY)
+            {
+#if PY_VERSION_HEX >= 0x02060000
+                return PyByteArray_FromStringAndSize(buffer, bytesUsed);
+#else
+                return PyBytes_FromStringAndSize(buffer, bytesUsed);
+#endif
+            }
+
+            if (sizeof(SQLWCHAR) == Py_UNICODE_SIZE)
                 return PyUnicode_FromUnicode((const Py_UNICODE*)buffer, bytesUsed / element_size);
 
-            return PyUnicode_FromWideChar((const wchar_t*)buffer, bytesUsed / element_size);
+            return PyUnicode_FromSQLWCHAR((const SQLWCHAR*)buffer, bytesUsed / element_size);
         }
 
-        if (PyString_CheckExact(bufferOwner))
-        {
-            if (_PyString_Resize(&bufferOwner, bytesUsed) == -1)
-                return 0;
-            PyObject* tmp = bufferOwner;
-            bufferOwner = 0;
-            buffer      = 0;
-            return tmp;
-        }
-
-        if (PyUnicode_CheckExact(bufferOwner))
+        if (bufferOwner && PyUnicode_CheckExact(bufferOwner))
         {
             if (PyUnicode_Resize(&bufferOwner, bytesUsed / element_size) == -1)
                 return 0;
@@ -205,25 +231,51 @@ public:
             return tmp;
         }
 
-        // We have allocated our own wchar_t buffer and must now copy it to a Unicode object.
-        PyObject* result = PyUnicode_FromWideChar((const wchar_t*)buffer, bytesUsed / element_size);
+        if (bufferOwner && PyBytes_CheckExact(bufferOwner))
+        {
+            if (_PyBytes_Resize(&bufferOwner, bytesUsed) == -1)
+                return 0;
+            PyObject* tmp = bufferOwner;
+            bufferOwner = 0;
+            buffer      = 0;
+            return tmp;
+        }
+
+#if PY_VERSION_HEX >= 0x02060000
+        if (bufferOwner && PyByteArray_CheckExact(bufferOwner))
+        {
+            if (PyByteArray_Resize(bufferOwner, bytesUsed) == -1)
+                return 0;
+            PyObject* tmp = bufferOwner;
+            bufferOwner = 0;
+            buffer      = 0;
+            return tmp;
+        }
+#endif
+
+        // We have allocated our own SQLWCHAR buffer and must now copy it to a Unicode object.
+        I(bufferOwner == 0);
+        PyObject* result = PyUnicode_FromSQLWCHAR((const SQLWCHAR*)buffer, bytesUsed / element_size);
         if (result == 0)
             return false;
-        free(buffer);
+        pyodbc_free(buffer);
         buffer = 0;
         return result;
     }
 };
 
-static PyObject*
-GetDataString(Cursor* cur, int iCol)
-{
-    // Returns a String or Unicode object for character and binary data.
 
+static PyObject* GetDataString(Cursor* cur, Py_ssize_t iCol)
+{
+    // Returns a string, unicode, or bytearray object for character and binary data.
+    //
+    // In Python 2.6+, binary data is returned as a byte array.  Earlier versions will return an ASCII str object here
+    // which will be wrapped in a buffer object by the caller.
+    //
     // NULL terminator notes:
     //
     //  * pinfo->column_size, from SQLDescribeCol, does not include a NULL terminator.  For example, column_size for a
-    //    char(10) column would be 10.  (Also, when dealing with wchar_t, it is the number of *characters*, not bytes.)
+    //    char(10) column would be 10.  (Also, when dealing with SQLWCHAR, it is the number of *characters*, not bytes.)
     //
     //  * When passing a length to PyString_FromStringAndSize and similar Unicode functions, do not add the NULL
     //    terminator -- it will be added automatically.  See objects/stringobject.c
@@ -253,10 +305,15 @@ GetDataString(Cursor* cur, int iCol)
     case SQL_LONGVARCHAR:
     case SQL_GUID:
     case SQL_SS_XML:
+#if PY_MAJOR_VERSION < 3
         if (cur->cnxn->unicode_results)
             nTargetType  = SQL_C_WCHAR;
         else
             nTargetType  = SQL_C_CHAR;
+#else
+        nTargetType  = SQL_C_WCHAR;
+#endif
+
         break;
 
     case SQL_WCHAR:
@@ -264,7 +321,7 @@ GetDataString(Cursor* cur, int iCol)
     case SQL_WLONGVARCHAR:
         nTargetType  = SQL_C_WCHAR;
         break;
-        
+
     default:
         nTargetType  = SQL_C_BINARY;
         break;
@@ -279,11 +336,22 @@ GetDataString(Cursor* cur, int iCol)
         SQLLEN cbData = 0;
 
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, buffer.GetBuffer(), buffer.GetRemaining(), &cbData);
+        ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), nTargetType, buffer.GetBuffer(), buffer.GetRemaining(), &cbData);
         Py_END_ALLOW_THREADS;
 
-        if (cbData == SQL_NULL_DATA)
+        if (cbData == SQL_NULL_DATA || (ret == SQL_SUCCESS && cbData < 0))
+        {
+            // HACK: FreeTDS 0.91 on OS/X returns -4 for NULL data instead of SQL_NULL_DATA (-1).  I've traced into the
+            // code and it appears to be the result of assigning -1 to a SQLLEN:
+            //
+            //   if (colinfo->column_cur_size < 0) {
+            //       /* TODO check what should happen if pcbValue was NULL */
+            //       *pcbValue = SQL_NULL_DATA;
+            //
+            // I believe it will be fine to treat all negative values as NULL for now.
+
             Py_RETURN_NONE;
+        }
 
         if (!SQL_SUCCEEDED(ret) && ret != SQL_NO_DATA)
             return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
@@ -332,7 +400,7 @@ GetDataString(Cursor* cur, int iCol)
             // For some reason, the NULL terminator is used in intermediate buffers but not in this final one.
             buffer.AddUsed(cbData);
         }
-        
+
         if (ret == SQL_SUCCESS || ret == SQL_NO_DATA)
             return buffer.DetachValue();
     }
@@ -341,8 +409,24 @@ GetDataString(Cursor* cur, int iCol)
     return 0;
 }
 
-static PyObject*
-GetDataBuffer(Cursor* cur, Py_ssize_t iCol)
+
+static PyObject* GetDataUser(Cursor* cur, Py_ssize_t iCol, int conv)
+{
+    // conv
+    //   The index into the connection's user-defined conversions `conv_types`.
+
+    PyObject* value = GetDataString(cur, iCol);
+    if (value == 0)
+        return 0;
+
+    PyObject* result = PyObject_CallFunction(cur->cnxn->conv_funcs[conv], "(O)", value);
+    Py_DECREF(value);
+    return result;
+}
+
+
+#if PY_VERSION_HEX < 0x02060000
+static PyObject* GetDataBuffer(Cursor* cur, Py_ssize_t iCol)
 {
     PyObject* str = GetDataString(cur, iCol);
 
@@ -359,30 +443,31 @@ GetDataBuffer(Cursor* cur, Py_ssize_t iCol)
 
     return buffer;
 }
+#endif
 
-static PyObject*
-GetDataDecimal(Cursor* cur, int iCol)
+
+static PyObject* GetDataDecimal(Cursor* cur, Py_ssize_t iCol)
 {
-    // The SQL_NUMERIC_STRUCT support is hopeless (SQL Server ignores scale on input parameters and output columns), so
-    // we'll rely on the Decimal's string parsing.  Unfortunately, the Decimal author does not pay attention to the
-    // locale, so we have to modify the string ourselves.
+    // The SQL_NUMERIC_STRUCT support is hopeless (SQL Server ignores scale on input parameters and output columns,
+    // Oracle does something else weird, and many drivers don't support it at all), so we'll rely on the Decimal's
+    // string parsing.  Unfortunately, the Decimal author does not pay attention to the locale, so we have to modify
+    // the string ourselves.
     //
     // Oracle inserts group separators (commas in US, periods in some countries), so leave room for that too.
+    //
+    // Some databases support a 'money' type which also inserts currency symbols.  Since we don't want to keep track of
+    // all these, we'll ignore all characters we don't recognize.  We will look for digits, negative sign (which I hope
+    // is universal), and a decimal point ('.' or ',' usually).  We'll do everything as Unicode in case currencies,
+    // etc. are too far out.
 
-    ColumnInfo* pinfo = &cur->colinfos[iCol];
+    // TODO: Is Unicode a good idea for Python 2.7?  We need to know which drivers support Unicode.
 
-    SQLLEN cbNeeded = pinfo->column_size + 3 +      // sign, decimal, NULL
-                      (pinfo->column_size / 3) + 2; // grouping.  I believe this covers all cases.
-
-    SQLLEN cbFetched = 0;
-    char* sz = (char*)_alloca(cbNeeded);
-
-    if (sz == 0)
-        return PyErr_NoMemory();
+    SQLWCHAR buffer[100];
+    SQLLEN cbFetched = 0; // Note: will not include the NULL terminator.
 
     SQLRETURN ret;
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_CHAR, sz, cbNeeded, &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), SQL_C_WCHAR, buffer, sizeof(buffer), &cbFetched);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
@@ -390,38 +475,44 @@ GetDataDecimal(Cursor* cur, int iCol)
     if (cbFetched == SQL_NULL_DATA)
         Py_RETURN_NONE;
 
-    // The decimal class requires the decimal to be a period and does not allow thousands separators.  Clean it up.
+    // Remove non-digits and convert the databases decimal to a '.' (required by decimal ctor).
     //
-    // Unfortunately this code only handles single-character values, which might be good enough for decimals and
-    // separators, but is certainly not good enough for currency symbols.
-    //
-    // Note: cbFetched does not include the NULL terminator.
+    // We are assuming that the decimal point and digits fit within the size of SQLWCHAR.
 
-    for (int i = cbFetched - 1; i >=0; i--)
+    int cch = (int)(cbFetched / sizeof(SQLWCHAR));
+
+    for (int i = (cch - 1); i >= 0; i--)
     {
-        if (sz[i] == chGroupSeparator || sz[i] == '$' || sz[i] == chCurrencySymbol)
+        if (buffer[i] == chDecimal)
         {
-            memmove(&sz[i], &sz[i] + 1, cbFetched - i);
-            cbFetched--;
+            // Must force it to use '.' since the Decimal class doesn't pay attention to the locale.
+            buffer[i] = '.';
         }
-        else if (sz[i] == chDecimal)
+        else if ((buffer[i] < '0' || buffer[i] > '9') && buffer[i] != '-')
         {
-            sz[i] = '.';
+            memmove(&buffer[i], &buffer[i] + 1, (cch - i) * sizeof(SQLWCHAR));
+            cch--;
         }
     }
 
-    return PyObject_CallFunction(decimal_type, "s", sz);
+    I(buffer[cch] == 0);
+
+    Object str(PyUnicode_FromSQLWCHAR(buffer, cch));
+    if (!str)
+        return 0;
+
+    return PyObject_CallFunction(decimal_type, "O", str.Get());
 }
 
-static PyObject*
-GetDataBit(Cursor* cur, int iCol)
+
+static PyObject* GetDataBit(Cursor* cur, Py_ssize_t iCol)
 {
     SQLCHAR ch;
     SQLLEN cbFetched;
     SQLRETURN ret;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_BIT, &ch, sizeof(ch), &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), SQL_C_BIT, &ch, sizeof(ch), &cbFetched);
     Py_END_ALLOW_THREADS
 
     if (!SQL_SUCCEEDED(ret))
@@ -436,19 +527,19 @@ GetDataBit(Cursor* cur, int iCol)
     Py_RETURN_FALSE;
 }
 
-static PyObject*
-GetDataLong(Cursor* cur, int iCol)
+
+static PyObject* GetDataLong(Cursor* cur, Py_ssize_t iCol)
 {
     ColumnInfo* pinfo = &cur->colinfos[iCol];
 
-    SQLINTEGER value = 0;
-    SQLLEN cbFetched = 0;
+    SQLINTEGER value;
+    SQLLEN cbFetched;
     SQLRETURN ret;
 
     SQLSMALLINT nCType = pinfo->is_unsigned ? SQL_C_ULONG : SQL_C_LONG;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
@@ -462,19 +553,18 @@ GetDataLong(Cursor* cur, int iCol)
     return PyInt_FromLong(value);
 }
 
-static PyObject*
-GetDataLongLong(Cursor* cur, int iCol)
+
+static PyObject* GetDataLongLong(Cursor* cur, Py_ssize_t iCol)
 {
     ColumnInfo* pinfo = &cur->colinfos[iCol];
 
-    INT64 value = 0;
-    SQLLEN cbFetched = 0;
-    SQLRETURN ret;
-
     SQLSMALLINT nCType = pinfo->is_unsigned ? SQL_C_UBIGINT : SQL_C_SBIGINT;
+    SQLBIGINT   value;
+    SQLLEN      cbFetched;
+    SQLRETURN   ret;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched);
     Py_END_ALLOW_THREADS
 
     if (!SQL_SUCCEEDED(ret))
@@ -484,20 +574,20 @@ GetDataLongLong(Cursor* cur, int iCol)
         Py_RETURN_NONE;
 
     if (pinfo->is_unsigned)
-        return PyLong_FromLongLong(*(UINT64*)&value);
+        return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)(SQLUBIGINT)value);
 
-    return PyLong_FromLongLong(*(INT64*)&value);
+    return PyLong_FromLongLong((PY_LONG_LONG)value);
 }
 
-static PyObject*
-GetDataDouble(Cursor* cur, int iCol)
+
+static PyObject* GetDataDouble(Cursor* cur, Py_ssize_t iCol)
 {
     double value;
     SQLLEN cbFetched = 0;
     SQLRETURN ret;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_DOUBLE, &value, sizeof(value), &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), SQL_C_DOUBLE, &value, sizeof(value), &cbFetched);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
@@ -508,16 +598,16 @@ GetDataDouble(Cursor* cur, int iCol)
     return PyFloat_FromDouble(value);
 }
 
-static PyObject*
-GetSqlServerTime(Cursor* cur, int iCol)
+
+static PyObject* GetSqlServerTime(Cursor* cur, Py_ssize_t iCol)
 {
     SQL_SS_TIME2_STRUCT value;
-    
+
     SQLLEN cbFetched = 0;
     SQLRETURN ret;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_BINARY, &value, sizeof(value), &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), SQL_C_BINARY, &value, sizeof(value), &cbFetched);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
@@ -525,12 +615,12 @@ GetSqlServerTime(Cursor* cur, int iCol)
     if (cbFetched == SQL_NULL_DATA)
         Py_RETURN_NONE;
 
-    int micros = value.fraction / 1000; // nanos --> micros
+    int micros = (int)(value.fraction / 1000); // nanos --> micros
     return PyTime_FromTime(value.hour, value.minute, value.second, micros);
 }
 
-static PyObject*
-GetDataTimestamp(Cursor* cur, int iCol)
+
+static PyObject* GetDataTimestamp(Cursor* cur, Py_ssize_t iCol)
 {
     TIMESTAMP_STRUCT value;
 
@@ -538,7 +628,7 @@ GetDataTimestamp(Cursor* cur, int iCol)
     SQLRETURN ret;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_TYPE_TIMESTAMP, &value, sizeof(value), &cbFetched);
+    ret = SQLGetData(cur->hstmt, (SQLUSMALLINT)(iCol+1), SQL_C_TYPE_TIMESTAMP, &value, sizeof(value), &cbFetched);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
@@ -550,21 +640,32 @@ GetDataTimestamp(Cursor* cur, int iCol)
     {
     case SQL_TYPE_TIME:
     {
-        int micros = value.fraction / 1000; // nanos --> micros
+        int micros = (int)(value.fraction / 1000); // nanos --> micros
         return PyTime_FromTime(value.hour, value.minute, value.second, micros);
     }
-    
+
     case SQL_TYPE_DATE:
         return PyDate_FromDate(value.year, value.month, value.day);
     }
 
-    int micros = value.fraction / 1000; // nanos --> micros
+    int micros = (int)(value.fraction / 1000); // nanos --> micros
     return PyDateTime_FromDateAndTime(value.year, value.month, value.day, value.hour, value.minute, value.second, micros);
 }
 
 
-PyObject*
-GetData(Cursor* cur, Py_ssize_t iCol)
+int GetUserConvIndex(Cursor* cur, SQLSMALLINT sql_type)
+{
+    // If this sql type has a user-defined conversion, the index into the connection's `conv_funcs` array is returned.
+    // Otherwise -1 is returned.
+
+    for (int i = 0; i < cur->cnxn->conv_count; i++)
+        if (cur->cnxn->conv_types[i] == sql_type)
+            return i;
+    return -1;
+}
+
+
+PyObject* GetData(Cursor* cur, Py_ssize_t iCol)
 {
     // Returns an object representing the value in the row/field.  If 0 is returned, an exception has already been set.
     //
@@ -572,6 +673,12 @@ GetData(Cursor* cur, Py_ssize_t iCol)
 
     ColumnInfo* pinfo = &cur->colinfos[iCol];
 
+    // First see if there is a user-defined conversion.
+
+    int conv_index = GetUserConvIndex(cur, pinfo->sql_type);
+    if (conv_index != -1)
+        return GetDataUser(cur, iCol, conv_index);
+
     switch (pinfo->sql_type)
     {
     case SQL_WCHAR:
@@ -582,12 +689,19 @@ GetData(Cursor* cur, Py_ssize_t iCol)
     case SQL_LONGVARCHAR:
     case SQL_GUID:
     case SQL_SS_XML:
+#if PY_VERSION_HEX >= 0x02060000
+    case SQL_BINARY:
+    case SQL_VARBINARY:
+    case SQL_LONGVARBINARY:
+#endif
         return GetDataString(cur, iCol);
 
+#if PY_VERSION_HEX < 0x02060000
     case SQL_BINARY:
     case SQL_VARBINARY:
     case SQL_LONGVARBINARY:
         return GetDataBuffer(cur, iCol);
+#endif
 
     case SQL_DECIMAL:
     case SQL_NUMERIC:
